// CO2 measrement using CO2mini
// using libhdiapi (ports/comms/hidapi)
// 2017-01-07 Kouichi Hirabayashi
// 2022-08-23, ZGm27 support Kouichi Hirabayashi

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <ctype.h>
#include <math.h>
#include <hidapi/hidapi.h>
#ifdef __linux
#include <linux/uhid.h>
#include <linux/hiddev.h>
#endif

#define VENDOR_ID	0x04d9	// Holtek
#define PRODUCT_ID	0xa052	// USB-zyTemp
#define EP_IN		(0x1 | USB_ENDPOINT_IN)
#define EP_OUT		(0x1 | USB_ENDPOINT_OUT)

#define MAX_DEV		20

typedef unsigned char co2mon_data_t[8];

static void swap_char(unsigned char *, unsigned char *);
static void decode_buf(co2mon_data_t, co2mon_data_t, co2mon_data_t);
int meas(int, int *, int *, int *);
char *strsave(char *);
static void onalrm(int);
void dump(char *, int);
void error(char *, ...);
void Usage();

struct sigaction saalrm, osaalrm;

co2mon_data_t magic_table = { 0 };
co2mon_data_t data = { 0 };

struct usb_bus *bus;
char *Path[MAX_DEV];
int nDev;
hid_device *dev;
int Temp[MAX_DEV];	// room temperature * 10
int Co2[MAX_DEV];	// CO2 ppm
int Hum[MAX_DEV];	// Humidity
int list[MAX_DEV];	// measurement list

char *cmd;
int vflg;		// verbose

int main(int argc, char **argv)
{
  int i, k, n;
  char *s;
  wchar_t *t;
  struct hid_device_info *di, *p;

  cmd = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-')
    for (s = argv[0]+1; *s; s++)
	switch (*s) {
	case 'v':
		vflg++;
		break;
	default:
		Usage();
	}

  saalrm.sa_handler = onalrm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);

  p = di = hid_enumerate(VENDOR_ID, PRODUCT_ID);
  for (i = 0; p != NULL && i < MAX_DEV; i++) {
    Path[nDev++] = strsave(p->path);
#if 0
    if (vflg) {
	printf("#%d: ", i + 1);
	for (t = p->manufacturer_string; *t; t++)
		printf("%c", *t);
	printf("/");
	for (t = p->product_string; *t; t++)
		printf("%c", *t);
	printf("\n");
    }
#endif
	p = p->next;
  }
  hid_free_enumeration(di);
  if (nDev == 0)
	error("No CO2mini device found.");
  if (vflg)
	printf("found %d device(s)\n", nDev);

  for (i = 0; i < MAX_DEV; i++)
	list[i] = -1;
  if (argc == 0) {	// all devides
	n = nDev;
	for (i = 0; i < nDev; i++)
		list[i] = i;
  }
  else {	// selectd devices
	for (n = i = 0; i < argc; i++) {
		sscanf(argv[i], "%d", &k);
		if (k < 1 || nDev < k) {
			if (vflg)
				fprintf(stderr, "ignore %d\n", k);
			continue;	// ignore
		}
		list[n++] = --k;
	}
  }

  for (i = 0; i < n; i++) {
	if (list[i] < 0) {
		printf("-/-");
		continue;
	}
	meas(list[i], &Temp[i], &Co2[i], &Hum[i]);
	if (i > 0)
		printf(" ");
	if (Hum[i])
		printf("%.1f/%d/%.0f", (double)Temp[i] / 16.0 - 273.15,
			Co2[i], (double)Hum[i] * 0.01);
	else if (Co2[i])
		printf("%.1f/%d", (double)Temp[i] / 16.0 - 273.15, Co2[i]);
	else
		printf("-/-");
	if (vflg)
		printf("\n");
  }
  printf("\n");

  return 0;
}

int meas(int n, int *ptemp, int *pco2, int *hum)
{
  int r, w;

  co2mon_data_t magic_table = { 0 };
  co2mon_data_t data = { 0 };
  co2mon_data_t val;
  int m;
  unsigned char checksum;
  char buf[128];
  double Tamb = -100;
  int CntR = -100, RH = -100;

  if (!(dev = hid_open_path(Path[n])))
	error("hid_open: error");
  r = hid_send_feature_report(dev, magic_table, sizeof(co2mon_data_t));
  if (r < 0 || r != sizeof(co2mon_data_t)) {
        fprintf(stderr, "hid_send_feature_report: error\n");
        return 1;
  }

  do {
    m = hid_read_timeout(dev, data, sizeof(co2mon_data_t),
	5000 /* milliseconds */);
    if (m < 0) {
	fprintf(stderr, "hid_read_timeout: error\n");
    }
    if (m != sizeof(co2mon_data_t)) {
	fprintf(stderr,
		"hid_read_timeout: transferred %d bytes, expected %lu bytes\n",
			n, (unsigned long)sizeof(co2mon_data_t));
    }
    if (strncmp((char *)data + 4, "\r\0\0\0", 4) == 0)
	memcpy((char *)val, (char *)data, 8);  // New Model
    else
	decode_buf(val, data, magic_table);     // Old Model
    if (vflg)
	dump((char *)val, 8);

    if (val[4] != 0x0d)
    fprintf(stderr,
	"Unexpected data from device (data[4] = %02hhx, want 0x0d)\n",
		val[4]);
    checksum = (val[0] + val[1] + val[2]) & 0xff;
    if (vflg && checksum != val[3])
	fprintf(stderr, "checksum error (%02hhx, await %02hhx)\n",
		checksum, val[3]);

    w = (val[1] << 8) + val[2];

    switch (val[0]) {
    case 'B':
	Tamb = w;
	break;
    case 'P':
	CntR = w;
	break;
    case 'A':	// ZGm27
	RH = w;
	break;
    default:
	// ignore
	break;  
    }
  } while (Tamb == -100 || CntR == -100 || RH == -100);
  *ptemp = Tamb;
  *pco2 = CntR;
  *hum = RH;

  hid_close(dev);
  dev = NULL;
  return 0;
}

static void decode_buf(co2mon_data_t val, co2mon_data_t buf,
	co2mon_data_t magic_table)
{
  int i;
  const unsigned char magic_word[8] = "Htemp99e";
  unsigned char tmp;

  swap_char(&buf[0], &buf[2]);
  swap_char(&buf[1], &buf[4]);
  swap_char(&buf[3], &buf[7]);
  swap_char(&buf[5], &buf[6]);

  tmp = (buf[7] << 5);
  val[7] = (buf[6] << 5) | (buf[7] >> 3);
  val[6] = (buf[5] << 5) | (buf[6] >> 3);
  val[5] = (buf[4] << 5) | (buf[5] >> 3);
  val[4] = (buf[3] << 5) | (buf[4] >> 3);
  val[3] = (buf[2] << 5) | (buf[3] >> 3);
  val[2] = (buf[1] << 5) | (buf[2] >> 3);
  val[1] = (buf[0] << 5) | (buf[1] >> 3);
  val[0] = tmp | (buf[0] >> 3);

  for (i = 0; i < 8; ++i) {
        val[i] -= (magic_word[i] << 4) | (magic_word[i] >> 4);
  }

  if (vflg && val[4] != 0x0d) {
	fprintf(stderr,
		"Unexpected data from device (data[4] = %02hhx, want 0x0d)\n",
			val[4]);
  }
}

static void
swap_char(unsigned char *a, unsigned char *b)
{
    unsigned char tmp = *a;
    *a = *b;
    *b = tmp;
}

char *strsave(char *s)
{
  char *p;

  if ((p = (char *)malloc(strlen(s) + 1)) == NULL) {
	fprintf(stderr, "strsave(): mamory over.\n");
	exit(1);
  }
  return strcpy(p, s);
}

static void onalrm(int i)
{
  saalrm.sa_handler = onalrm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
  if (dev != NULL)
	hid_close(dev);
}

void dump(char *s, int n)
{
  int i;

  for (i = 0; i < n; i++) {
	printf(" %02x", s[i] & 0xff);
  }
  printf("\n");
}

void error(char *fmt, ...)
{
  va_list argp;

  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fprintf(stderr, "\n");
  exit(1);
}

void Usage()
{
  int i;
  char *f[] = {
"Commad: %s - Measurement of aerial CO2 concentration with CO2mini",
"Usage: %s [-v] [ n ..]",
"Option: -v		# verbose",
"Example:",
"  %s			# use all CO2mini only",
"  %s 1			# use first CO2mini only",
"  %s 2 4		# use 2'nd and 4'th CO2mini",
"Data Format: temp/CO2[/RH][ temp/CO2[/RH] ..]",
" print ambient temperature(C), CO2 concentration (ppm) and relative",
" humidity (ZGm27 only) for specfied device.",
NULL
};
  for (i = 0; f[i] != NULL; i++) {
	fprintf(stderr, f[i], cmd);
	fprintf(stderr, "\n");
  }
  exit(1);
}

