// CO2 measurement using CO2mini
// using libusb
// 2022.06.18 Kouichi Hirabayashi
// 2022.08.23 ZGm27 support, Kouichi Hirabayashi

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <err.h>
#ifdef __FreeBSD__
#include <libusb.h>
#elif __linux__
#include <libusb-1.0/libusb.h>
#include <linux/hiddev.h>
#endif


#define VENDOR_ID	0x04d9	// Holtek
#define PRODUCT_ID	0xa052	// USB-zyTemp

#define CTRL_OUT	0x21
#define HID_SET_REPORT	0x09
#ifndef __linux__
#define HID_REPORT_TYPE_FEATURE	0x03
#endif

#define TIMEOUT		20
#define MAX_DEV		20

int nDev;
libusb_device_handle *handle[MAX_DEV];
int seq[MAX_DEV];	// measurement seqence

typedef unsigned char co2mon_data_t[8];

static void swap_char(unsigned char *, unsigned char *);
static void decode_buf(co2mon_data_t, co2mon_data_t, co2mon_data_t);

int find_dev(int, int);
int meas(libusb_device_handle *);
char *strsave(char *);
static void onalrm(int);
void dump(char *, int);
void error(char *, ...);
void Usage();

struct sigaction saalrm, osaalrm;
co2mon_data_t val;

int Temp[MAX_DEV];	// room temperature * 10
int Co2[MAX_DEV];	// CO2 ppm

int vflg;
char *cmd;

int main(int argc, char **argv)
{
  int i, k, n;
  char *s;

  cmd = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-')
    for (s = argv[0]+1; *s; s++)
	switch (*s) {
	case 'v':
		vflg++;
		break;
	default:
	       Usage();
	}

  saalrm.sa_handler = onalrm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);

  for (i = 0; i < MAX_DEV; i++)
	seq[i] = -1;
  find_dev(VENDOR_ID, PRODUCT_ID);
  if (vflg)
	fprintf(stderr, "found %d device(s).\n", nDev);
  if (argc == 0) {
	for (n = 0; n < nDev; n++)
		seq[n] = n;
  }
  else {
    for (i = n = 0; i < argc; i++) {
	k = -1;
	sscanf(argv[i], "%d", &k);
	if (k < 1 || nDev < k) {
		if (vflg) {
			fprintf(stderr, "device #%d ignored.\n", k);
		}
		continue;
	}
	seq[n++] = k -1;
    }
  }
  for (i = 0; i < n; i++) {
	meas(handle[seq[i]]);
	if (i < n - 1)
		printf(" ");
	else
  		printf("\n");
	if (vflg)
		printf("\n");
  }
  for (i = 0; i < nDev; i++)
	libusb_close(handle[i]);
  libusb_exit(NULL);

  return 0;
}

void swap_char(unsigned char *a, unsigned char *b)
{
  unsigned char tmp = *a;

  *a = *b;
  *b = tmp;
}

void decode_buf(co2mon_data_t val, co2mon_data_t buf, co2mon_data_t magic_table)
{
  const unsigned char magic_word[8] = "Htemp99e";
  int i;
  unsigned char tmp;

  swap_char(&buf[0], &buf[2]);
  swap_char(&buf[1], &buf[4]);
  swap_char(&buf[3], &buf[7]);
  swap_char(&buf[5], &buf[6]);

  tmp = (buf[7] << 5);
  val[7] = (buf[6] << 5) | (buf[7] >> 3);
  val[6] = (buf[5] << 5) | (buf[6] >> 3);
  val[5] = (buf[4] << 5) | (buf[5] >> 3);
  val[4] = (buf[3] << 5) | (buf[4] >> 3);
  val[3] = (buf[2] << 5) | (buf[3] >> 3);
  val[2] = (buf[1] << 5) | (buf[2] >> 3);
  val[1] = (buf[0] << 5) | (buf[1] >> 3);
  val[0] = tmp | (buf[0] >> 3);
  for (i = 0; i < 8; ++i) {
	val[i] -= (magic_word[i] << 4) | (magic_word[i] >> 4);
  }
}

int find_dev(int vid, int pid) {
  libusb_device **list, *device;
  int i, n, m, r, vendorId, productId;
  struct libusb_device_descriptor desc;
  unsigned char text[512];

  libusb_init(NULL);
  n = libusb_get_device_list(NULL, &list);
  for (i = 0; i < n; i++) {
    device = list[i];
    libusb_get_device_descriptor(device, &desc);
    if (desc.idVendor == vid && desc.idProduct == pid) {
	r = libusb_open(device, &handle[nDev]);
	if (r != 0)
		fprintf(stderr, "libusb_open() error.\n");
	nDev++;
    }
  }
  libusb_free_device_list(list, 1);
  return nDev;
}

int meas(libusb_device_handle *devh)
{
  int Tamb = -100, CntR = -100, RH = -100;
  co2mon_data_t magic_table = { 0 };
  co2mon_data_t data = { 0 };
  int k, n, m, cksum, r, w;

  alarm(TIMEOUT);
#ifndef __linux__
  r = libusb_set_configuration(devh, 1);
  if (r)
	fprintf(stderr, "libusb_set_configuration() error.\n");
#endif
  r = libusb_claim_interface(devh, 0);
  if (r)
	fprintf(stderr, "libusb_claim_interface() error.\n");
  r = libusb_control_transfer(devh, CTRL_OUT, HID_SET_REPORT,
	(HID_REPORT_TYPE_FEATURE<<8) | 0x00, 0, data, sizeof(magic_table),
		TIMEOUT);
  if (r != sizeof(magic_table))
	fprintf(stderr, "libusb_control_transfer() error.\n");
  do {
#ifndef INTERRUPT_TRANSFER 
    r = libusb_bulk_transfer(devh, LIBUSB_ENDPOINT_IN|1,
	data, sizeof(data), &m, 4000);
#else
    r = usb_interrupt_read((usb_dev_handle *)devh, LIBUSB_ENDPOINT_IN|1,
	(char *)data, sizeof(data), 4000);
printf("r(%d)\n", r);
#endif
//    dump((char *)data, m);
    if (strncmp((char *)data + 4, "\r\0\0\0", 4) == 0)
	memcpy((char *)val, (char *)data, 8);	// New Model
    else
	decode_buf(val, data, magic_table);	// Old Model
    if (vflg)
	dump((char *)val, 8);
    if (val[4] != 0x0D)
		fprintf(stderr, "Unexpected data (%02X) != 0x0D.\n", val[4]);
    cksum = (val[0] + val[1] + val[2]) & 0xFF;
    if (cksum != val[3])
	fprintf(stderr, "checksum error (%02X) != %02X\n",
			cksum, val[3]);
    w = (val[1] <<8) + val[2];
    switch (val[0]) {
    case 'B':	// 0x42 Temp (Ambient Temperature)
	Tamb = w;
	break;
    case 'P':	// 0x50 CntR (Relative Concentration of CO2)
	CntR = w;
	break;
    case 'A':	// ZGm27
	RH = w;
	break;
    default:
	// ignore
	break;
    }
  } while (Tamb == -100 || CntR == -100 || RH == -100);
  alarm(0);
  Temp[nDev] = Tamb;
  Co2[nDev] = CntR;
  if (RH != 0)
    printf("%.1f/%d/%.0f", (double)Tamb / 16.0 - 273.15, CntR, RH * 1e-2);
  else
    printf("%.1f/%d", (double)Tamb / 16.0 - 273.15, CntR);
  return 0;
}

static void onalrm(int i)
{
  saalrm.sa_handler = onalrm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
}

void dump(s, n) char *s; int n;
{
  int i;

  for (i = 0; i < n; i++) {
	printf(" %02x", s[i] & 0xff);
  }
  printf("\n");
}

void error(char *fmt, ...)
{
  va_list argp;

  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fprintf(stderr, "\n");
  exit(1);
}

void Usage()
{
  static char *t[] = {
"Commad %s - CO2mini access program",
"Usage	%s [n ..]",
"Example	%s	# using all CO2mini device",
"		%s 2	# using second CO2mini device",
"		%s 1 3	# using first and third CO2mini device",
"Data Format: temp/CO2[RH][ temp/CO2[/RH] ..]",
" print ambient temperature(C), CO2 concentration (ppm) and relative",
" humidity (ZGm27 only) for specfied device.",
  NULL
};
  int i;

  for (i = 0; t[i] != NULL; i++) {
	fprintf(stderr, t[i], cmd);
	fprintf(stderr, "\n");
  }
  exit(1);
}

